import argparse
import dataclasses
import inspect
from argparse import _ActionsContainer
from logging import getLogger
from typing import Any, Dict, List, Optional, Set, Tuple, Type, Union

from .. import utils
from . import docstring
from .wrapper import Wrapper

logger = getLogger(__name__)


class FieldWrapper(Wrapper[dataclasses.Field]):
    """
    FieldWrapper represents a single "atomic" field of a dataclass. That is, a field
    that does not have any subfields (that we attempt to parse).
    The FieldWrapper is essentially a reified call to
     the `add_argument(*option_strings, **arg_options)` function of the
    `argparse._ArgumentGroup`. Typically this is called by DataclassWrapper.

    The `option_strings`, `required`, `help`, `default`, etc.
    attributes just autogenerate the argument of the same name of the
    above-mentioned `add_argument` function. The `arg_options` attribute fills
    in the rest and may overwrite these values, depending on the type of field.

    The `field` argument is the actually wrapped `dataclasses.Field` instance.
    """

    @property
    def field(self) -> dataclasses.Field:
        return self._field

    def __init__(
        self,
        field: dataclasses.Field,
        parent: Optional[Wrapper] = None,
        preferred_help: str = docstring.HelpOrder.inline,
    ):
        self._field = field
        self._parent: Optional[Wrapper] = parent
        # Holders used to 'cache' the properties.
        # (could've used cached_property with Python 3.8).
        self._option_strings: Optional[Set[str]] = None
        self._required: Optional[bool] = None
        self._docstring: docstring.AttributeDocString = docstring.AttributeDocString()
        self._help: Optional[str] = None
        self._default: Optional[Union[Any, List[Any]]] = None
        self._dest: Optional[str] = None
        # the argparse-related options:
        self._arg_options: Dict[str, Any] = {}
        self._type: Optional[Type[Any]] = None

        # preferred parse for docstring / help text in < inline | above | below >
        self.preferred_help = preferred_help

        # stores the resulting values for each of the destination attributes.
        self._results: Dict[str, Any] = {}

    @property
    def arg_options(self) -> Dict[str, Any]:
        """Dictionary of values to be passed to the `add_argument` method.

        The main feature of this package is to infer these arguments
        automatically using features of the built-in `dataclasses` package, as
        well as Python's type annotations.

        By passing additional keyword arguments to the `field()`
        function, the autogenerated arguments can be overwritten,
        giving access to all of the usual argparse features know and love.

        NOTE: When passing an `action` keyword argument, we remove all the
        autogenerated options that aren't required by the Action class
        constructor.
        For example, when specifying a custom `action` like "store_true" or
        "store_false", the `type` argument autogenerated here shouldn't be
        passed to the constructor of the `argparse._StoreFalseAction`, so we
        discard it.
        """
        if self._arg_options:
            return self._arg_options
        # get the auto-generated options.
        options = self.get_arg_options()
        # overwrite the auto-generated options with given ones, if any.
        options.update(self.custom_arg_options)
        # only keep the arguments used by the Action constructor.
        action = options.get("action", "store")
        self._arg_options = only_keep_action_args(options, action)
        return self._arg_options

    def get_arg_options(self) -> Dict[str, Any]:
        """Create the `parser.add_arguments` kwargs for this field."""
        if not self.field.init:
            return {}
        # TODO: Refactor this:
        # 1. Create a `get_argparse_options_for_field` function
        # 2. Use `get_argparse_options_for_annotation` below as part of that function
        # 3. Update the dict returned from 1. with values set in the field() function
        # 4. Update the dict from 3. with the values set by the DataclassWrapper, or
        # when this field is reused. (are they ever modified externally?)
        # 5. Return that dictionary.
        _arg_options: Dict[str, Any] = {}

        _arg_options["required"] = False  # Required arguments can also be set from yaml,
        # so do not enforce with argparse
        _arg_options["dest"] = self.dest
        _arg_options["default"] = self.default

        if self.help:
            _arg_options["help"] = self.help
        elif self.default is not None:
            # issue 64: Need to add an empty 'help' string, so that the formatter
            # automatically adds the (default: '123')
            _arg_options["help"] = " "

        tpe = self.type

        tpe = utils.canonicalize_union(tpe)

        _arg_options["type"] = tpe
        try:
            _arg_options["type"].__name__ = self.type.__repr__().replace("typing.", "")
        except Exception:
            # Only to prettify printing, if fails just continue
            pass

        return _arg_options

    @property
    def action(self) -> Union[str, Type[argparse.Action]]:
        """The `action` argument to be passed to `add_argument(...)`."""
        return self.custom_arg_options.get("action", "store")

    @property
    def action_str(self) -> str:
        if isinstance(self.action, str):
            return self.action
        return self.action.__name__

    @property
    def custom_arg_options(self) -> Dict[str, Any]:
        """Custom argparse options that overwrite those in `arg_options`.

        Can be set by using the `field` function, passing in a keyword argument
        that would usually be passed to the parser.add_argument(
        *option_strings, **kwargs) method.
        """
        return self.field.metadata.get("custom_args", {})

    @property
    def option_strings(self) -> List[str]:
        """Generates the `option_strings` argument to the `add_argument` call.

        `parser.add_argument(*name_or_flags, **arg_options)`

        ## Notes:
        - Additional names for the same argument can be added via the `field`
        function.
        - Whenever the name of an attribute includes underscores ("_"), the same
        argument can be passed by using dashes ("-") instead. This also includes
        aliases.
        - If an alias contained leading dashes, either single or double, the
        same number of dashes will be used

        For an illustration of this, see the aliases example.

        """

        dashes: List[str] = []  # contains the leading dashes.
        options: List[str] = []  # contains the name following the dashes.

        # Currently create only a single option name, no support for aliases
        dashes.append("--")
        options.append(self.dest)

        # remove duplicates by creating a set.
        option_strings = set(f"{dash}{option}" for dash, option in zip(dashes, options))
        return list(sorted(option_strings, key=len))

    @property
    def nargs(self):
        return self.custom_arg_options.get("nargs", None)

    @property
    def default(self) -> Any:
        """Either a single default value, when parsing a single argument, or
        the list of default values, when this argument is reused multiple times
        (which only happens with the `ConflictResolution.ALWAYS_MERGE` option).

        In order of increasing priority, this could either be:
        1. The default attribute of the field
        2. the value of the corresponding attribute on the parent,
        if it has a default value
        """
        if self._default is not None:
            return self._default

        default: Any = utils.default_value(self.field)

        if default is dataclasses.MISSING:
            default = None

        self._default = default
        return self._default

    @default.setter
    def default(self, value: Any):
        self._default = value

    @property
    def required(self) -> bool:
        if self._required is not None:
            return self._required

        if self.action_str.startswith("store_"):
            # all the store_* actions do not require a value.
            self._required = False
        elif self.is_optional:
            self._required = False
        elif self.parent.required:
            # if the parent dataclass is required, then this attribute is too.
            # TODO: does that make sense though?
            self._required = True

        elif self.nargs in {"?", "*"}:
            self._required = False
        elif self.nargs == "+":
            self._required = True

        elif self.default is None:
            self._required = True
        else:
            self._required = False
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value

    @property
    def type(self) -> Type[Any]:
        """Returns the wrapped field's type annotation."""
        if self._type is None:
            self._type = self.field.type

        return self._type

    def __str__(self):
        return f"""<FieldWrapper for field '{self.dest}'>"""

    @property
    def help(self) -> Optional[str]:
        if self._help:
            return self._help
        try:
            self._docstring = docstring.get_attribute_docstring(self.parent.dataclass, self.field.name)
        except (SystemExit, Exception) as e:
            logger.debug(f"Couldn't find attribute docstring for field {self.name}, {e}")
            self._docstring = docstring.AttributeDocString()

        help_text = docstring.get_preferred_help_text(self._docstring, preferred_help=self.preferred_help)
        if help_text is not None:
            self._help = help_text
        return self._help

    @help.setter
    def help(self, value: str):
        self._help = value

    @property
    def name(self) -> str:
        return self.field.name

    @property
    def is_list(self):
        return utils.is_list(self.type)

    @property
    def is_enum(self) -> bool:
        return utils.is_enum(self.type)

    @property
    def is_tuple(self) -> bool:
        return utils.is_tuple(self.type)

    @property
    def is_bool(self) -> bool:
        return utils.is_bool(self.type)

    @property
    def is_optional(self) -> bool:
        return utils.is_optional(self.field.type)

    @property
    def is_union(self) -> bool:
        return utils.is_union(self.field.type)

    @property
    def type_arguments(self) -> Optional[Tuple[Type, ...]]:
        return utils.get_type_arguments(self.type)

    @property
    def parent(self):
        return self._parent

    def add_action(self, parser: _ActionsContainer) -> None:
        logger.debug(f"Arg options for field '{self.name}': {self.arg_options}")
        parser.add_argument(*self.option_strings, **self.arg_options)


def only_keep_action_args(options: Dict[str, Any], action: Union[str, Any]) -> Dict[str, Any]:
    """Remove all the arguments in `options` that aren't required by the Action.

    Parameters
    ----------
    options : Dict[str, Any]
        A dictionary of options that would usually be passed to
        `add_arguments(*option_strings, **options)`.
    action : Union[str, Any]
        The action class or name.

    Returns
    -------
    Dict[str, Any]
        [description]
    """
    # TODO: explicitly tests these custom actions?
    argparse_action_classes: Dict[str, Type[argparse.Action]] = {
        "store": argparse._StoreAction,
        "store_const": argparse._StoreConstAction,
        "store_true": argparse._StoreTrueAction,
        "store_false": argparse._StoreFalseAction,
        "append": argparse._AppendAction,
        "append_const": argparse._AppendConstAction,
        "count": argparse._CountAction,
        "help": argparse._HelpAction,
        "version": argparse._VersionAction,
        "parsers": argparse._SubParsersAction,
    }
    if action not in argparse_action_classes:
        # the provided `action` is not a standard argparse-action.
        # We don't remove any of the provided options.
        return options

    # Remove all the keys that aren't needed by the action constructor:
    action_class = argparse_action_classes[action]
    argspec = inspect.getfullargspec(action_class)

    if argspec.varargs is not None or argspec.varkw is not None:
        # if the constructor takes variable arguments, pass all the options.
        logger.debug("Constructor takes var args. returning all options.")
        return options

    args_to_keep = [*argspec.args, "action"]

    kept_options, deleted_options = utils.keep_keys(options, args_to_keep)
    if deleted_options:
        logger.debug(
            "Some auto-generated options were deleted, as they were "
            f"not required by the Action constructor: {deleted_options}."
        )
        logger.debug(f"Kept options: \t{kept_options.keys()}")
        logger.debug(f"Removed options: \t{deleted_options.keys()}")
    return kept_options



from dataclasses import dataclass
from typing import Tuple

import pytest

from draccus import ParsingError
from draccus.utils import DecodingError
from tests.testutils import TestSetup


def test_tuple_with_n_items_takes_only_n_values():
    @dataclass
    class Container(TestSetup):
        ints: Tuple[int, int] = (1, 5)

    c = Container.setup("")
    assert c.ints == (1, 5)
    c = Container.setup("--ints [4,8]")
    assert c.ints == (4, 8)
    with pytest.raises(DecodingError) as e:
        c = Container.setup("--ints [4,5,6,7,8]")

    assert "Expected 2 items, got 5" in str(e.value)
    assert e.value.key_path == ("ints",)


def test_tuple_elipsis_takes_any_number_of_args():
    @dataclass
    class Container(TestSetup):
        ints: Tuple[int, ...] = (1, 2, 3)

    c = Container.setup("")
    assert c.ints == (1, 2, 3)
    c = Container.setup("--ints '[4, 5, 6, 7, 8]'")
    assert c.ints == (4, 5, 6, 7, 8)


def test_each_type_is_used_correctly():
    @dataclass
    class Container(TestSetup):
        """A container with mixed items in a tuple."""

        mixed: Tuple[int, str, bool, float] = (1, "bob", False, 1.23)

    c = Container.setup("")
    assert c.mixed == (1, "bob", False, 1.23)

    c = Container.setup("--mixed [1,2,false,1]")
    assert c.mixed == (1, "2", False, 1.0)



import json
from dataclasses import dataclass, field
from enum import Enum, auto
from typing import Dict, List, Optional, Tuple

import yaml

import draccus
from draccus.utils import DraccusException

from .testutils import *


class Color(Enum):
    blue = auto()
    red = auto()


def test_encode_something(simple_attribute):
    some_type, _, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        d: Dict[str, some_type] = field(default_factory=dict)
        l: List[Tuple[some_type, some_type]] = field(default_factory=list)
        t: Dict[str, Optional[some_type]] = field(default_factory=dict)

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})

    assert draccus.decode(SomeClass, draccus.encode(b)) == b


@parametrize("config_type", ["", "yaml", "json", "toml"])
def test_dump_load(simple_attribute, config_type, tmp_path):
    some_type, _, expected_value = simple_attribute

    if config_type != "":
        draccus.set_config_type(config_type)

    @dataclass
    class SomeClass:
        val: Optional[some_type] = None

    b = SomeClass(val=expected_value)

    tmp_file = tmp_path / "config"
    draccus.dump(b, tmp_file.open("w"))

    new_b = draccus.parse(config_class=SomeClass, config_path=tmp_file, args="")
    assert new_b == b

    arguments = shlex.split(f"--config_path {tmp_file}")
    new_b = draccus.parse(config_class=SomeClass, args=arguments)
    assert new_b == b

    new_b = draccus.parse(config_class=SomeClass, args="")
    assert new_b != b

    draccus.set_config_type("yaml")


def test_dump_load_context():
    @dataclass
    class SomeClass:
        val: str = "hello"

    b = SomeClass()

    yaml_str = draccus.dump(b)
    assert yaml_str == yaml.dump(draccus.encode(b))

    with draccus.config_type("json"):
        json_str = draccus.dump(b)
        assert json_str == json.dumps(draccus.encode(b))

    yaml_str = draccus.dump(b)
    assert yaml_str == yaml.dump(draccus.encode(b))

    assert draccus.get_config_type() is draccus.ConfigType.YAML


def test_dump_load_dicts(simple_attribute, tmp_path):
    some_type, _, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        d: Dict[str, some_type] = field(default_factory=dict)
        l: List[Tuple[some_type, some_type]] = field(default_factory=list)
        t: Dict[str, Optional[some_type]] = field(default_factory=dict)

    b = SomeClass()
    b.d.update({"hey": expected_value})
    b.l.append((expected_value, expected_value))
    b.t.update({"hey": None, "hey2": expected_value})

    tmp_file = tmp_path / "config"
    draccus.dump(b, tmp_file.open("w"))

    new_b = draccus.parse(config_class=SomeClass, config_path=tmp_file, args="")
    assert new_b == b
    arguments = shlex.split(f"--config_path {tmp_file}")
    new_b = draccus.parse(config_class=SomeClass, args=arguments)
    assert new_b == b


def test_dump_load_enum(tmp_path):
    @dataclass
    class SomeClass:
        color: Color = Color.red

    b = SomeClass()
    tmp_file = tmp_path / "config.yaml"
    draccus.dump(b, tmp_file.open("w"))

    new_b = draccus.parse(config_class=SomeClass, config_path=tmp_file, args="")
    assert new_b == b


def test_reserved_config_word():
    @dataclass
    class MainClass:
        config_path: str = ""

    with raises(DraccusException):
        draccus.parse(MainClass)


def test_super_nesting():
    @dataclass
    class Complicated:
        x: List[List[List[Dict[int, Tuple[int, float, str, List[float]]]]]] = field(default_factory=list)

    c = Complicated()
    c.x = [[[{0: (2, 1.23, "bob", [1.2, 1.3])}]]]
    assert draccus.decode(Complicated, draccus.encode(c)) == c



__version__ = "0.8.0"

from . import utils, wrappers
from .argparsing import parse, wrap
from .cfgparsing import dump, load
from .choice_types import CHOICE_TYPE_KEY, ChoiceRegistry, ChoiceType, PluginRegistry
from .fields import field
from .options import ConfigType, Options, config_type
from .parsers.decoding import decode
from .parsers.encoding import encode
from .utils import ParsingError

get_config_type = Options.get_config_type
set_config_type = Options.set_config_type



import dataclasses

import pytest

from draccus import ParsingError
from draccus.choice_types import ChoiceRegistry

from .testutils import TestSetup


@dataclasses.dataclass
class Person(ChoiceRegistry):
    name: str  # Person's name


@dataclasses.dataclass
class Adult(Person):
    age: int


@dataclasses.dataclass
class Child(Person):
    favorite_toy: str  # Child's favorite toy


Person.register_subclass("adult", Adult)
Person.register_subclass("child", Child)


def test_choice_registry_argparse():
    @dataclasses.dataclass
    class Something(TestSetup):
        person: Person = Adult("bob", 10)

    s = Something.setup("")
    assert s.person == Adult("bob", 10)

    s = Something.setup("--person.type child --person.name bob --person.favorite_toy truck")
    assert s.person == Child("bob", "truck")

    with pytest.raises(SystemExit):
        Something.setup("--person.type baby --person.name bob")

    with pytest.raises(ParsingError):
        Something.setup("--person.type adult --person.name bob --person.age 10 --person.favorite_toy truck")

    with pytest.raises(ParsingError):
        Something.setup("--person.name hi")


@dataclasses.dataclass
class Something(TestSetup):
    person: Person = Adult("bob", 10)


def test_choice_registry_examine_help():
    # TODO: why is the default: None here?
    target = """
    usage: draccus [-h] [--config_path str] [--person str]
               [--person.type {adult,child}] [--person.age int]
               [--person.name str] [--person.favorite_toy str]

options:
  -h, --help            show this help message and exit
  --config_path str     Path for a config file to parse with draccus (default:
                        None)
  --person str          Config file for person (default: None)

Something:

Person ['person']:

  --person.type {adult,child}
                        Which type of Person ['person'] to use (default: None)

Adult ['person']:

  --person.name str     Person's name (default: None)
  --person.age int

Child ['person']:

  --person.name str     Person's name (default: None)
  --person.favorite_toy str
                        Child's favorite toy (default: None)
"""
    print(Something.get_help_text().strip())
    assert Something.get_help_text().strip() == target.strip()



import tempfile
from enum import Enum, auto
from pathlib import Path

from tests.conftest import Optimizers


class Color(Enum):
    blue: str = auto()  # type: ignore
    red: str = auto()  # type: ignore


def test_load_hyperparameters_with_include(HyperParameters):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        config_path = tmpdir / "config.yaml"
        config_path.write_text(
            """
age_group: !include ./age_group.yaml
            """
        )

        age_group_path = tmpdir / "age_group.yaml"
        age_group_path.write_text(
            """
name: age_group
num_units: 16
"""
        )

        config = HyperParameters.setup(f"--config_path {config_path}")

        assert config.age_group.name == "age_group"
        assert config.age_group.num_units == 16


def test_merge_include_hyperparameters(HyperParameters):
    with tempfile.TemporaryDirectory() as tmpdir:
        tmpdir = Path(tmpdir)

        config_path = tmpdir / "config.yaml"
        config_path.write_text(
            """
<<: !include ./base_config.yaml
age_group: !include ./age_group.yaml
embedding_dim: 32
            """,
        )

        base_config_path = tmpdir / "base_config.yaml"
        base_config_path.write_text(
            """
batch_size: 13
optimizer: ADAM
embedding_dim: 16
"""
        )

        age_group_path = tmpdir / "age_group.yaml"
        age_group_path.write_text(
            """
name: age_group
num_units: 16
"""
        )

        config = HyperParameters.setup(f"--config_path {config_path}")

        assert config.batch_size == 13
        assert config.optimizer == Optimizers.ADAM
        assert config.embedding_dim == 32
        assert config.age_group.name == "age_group"
        assert config.age_group.num_units == 16



"""Simple, extendable mechanism for encoding pracitaclly anything to string.

Just register a new encoder for a given type like so:

import draccus
import numpy as np
@draccus.encode.register
def encode_ndarray(obj: np.ndarray) -> str:
    return obj.tostring()
"""
import json
import typing
from argparse import Namespace
from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from enum import Enum
from functools import singledispatch
from logging import getLogger
from os import PathLike
from typing import Any, Dict, Hashable, List, Tuple, Union

from draccus.choice_types import CHOICE_TYPE_KEY, ChoiceType

logger = getLogger(__name__)


class SimpleJsonEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        return encode(o)


"""
# NOTE: This code is commented because of static typing check error.
# The problem is incompatibility of mypy and singledispatch.
# See mypy issues for more info:
# https://github.com/python/mypy/issues/8356
# https://github.com/python/mypy/issues/2904
# https://github.com/python/mypy/issues/9112#issuecomment-725316936

class Dataclass(Protocol):
    # see dataclasses.is_dataclass implementation with _FIELDS
    __dataclass_fields__: Dict[str, Field[Any]]


T = TypeVar("T", bool, int, None, str)


@overload
def encode(obj: Dataclass) -> Dict[str, Any]: ...

@overload
def encode(obj: Union[List[Any], Set[Any], Tuple[Any, ...]]) -> List[Any]:
    ...

@overload
def encode(obj: Mapping[Any, Any]) -> Dict[Any, Any]: ...

@overload
def encode(obj: T) -> T: ...
"""


@singledispatch
def encode(obj: Any) -> Any:
    """Encodes an object into a json/yaml-compatible primitive type."""
    try:
        if isinstance(obj, ChoiceType):
            return encode_choice(obj)
        elif is_dataclass(obj):
            return encode_dataclass(obj)
        elif obj is None:
            return None
        else:
            raise Exception(f"No parser for object {obj} of type {type(obj)}, consider using draccus.encode.register")
    except Exception as e:
        logger.debug(f"Cannot encode object {obj}: {e}")
        raise e


def encode_dataclass(obj: Any):
    d: Dict[str, Any] = dict()
    for field in fields(obj):
        value = getattr(obj, field.name)
        try:
            d[field.name] = encode(value)
        except TypeError as e:
            logger.error(f"Unable to encode field {field.name}: {e}")
            raise e
    return d


def encode_choice(obj: ChoiceType):
    encoded = encode_dataclass(obj)

    if not isinstance(encoded, dict):
        raise Exception(f"Choice Class {obj} is not encoded as a dict: {encoded}")

    encoded = {CHOICE_TYPE_KEY: obj.get_choice_name(type(obj)), **encoded}

    return encoded


@encode.register(Mapping)
def encode_dict(obj: Mapping) -> Union[typing.Mapping[Any, Any], List[Tuple[Any, Any]]]:
    constructor = type(obj)
    result: Union[Mapping, List[Tuple[Any, Any]]] = constructor()
    for k, v in obj.items():
        k_ = encode(k)
        v_ = encode(v)
        if isinstance(result, list):
            result.append((k_, v_))
        elif isinstance(k_, Hashable):
            result[k_] = v_  # type: ignore
        else:
            # If the encoded key isn't "Hashable", then we store it as a list of tuples
            if isinstance(result, Mapping):
                result = list(result.items())
            result.append((k_, v_))
    return result


@encode.register(Enum)
def encode_enum(obj: Enum) -> str:
    return obj.name


for t in [str, float, int, bool, bytes]:
    encode.register(t, lambda x: x)

for t in [list, tuple, set]:
    encode.register(t, lambda x: list(map(encode, x)))

encode.register(PathLike, lambda x: x.__fspath__())

encode.register(Namespace, lambda x: encode(vars(x)))



"""Abstract Wrapper base-class for the FieldWrapper and DataclassWrapper."""
import argparse
from abc import ABC, abstractmethod
from dataclasses import Field
from typing import Generic, List, Optional, Type

from draccus.utils import T

# We can think of a Wrapper as a node in a tree, where the root is the DataclassWrapper for the root dataclass, and the
# leaves are the FieldWrappers. (So internal nodes are DataclassWrappers, for now.)


class Wrapper(Generic[T], ABC):
    @property
    def dest(self) -> str:
        """Where the attribute will be stored in the Namespace."""
        if self.parent is None:
            return self.name
        lineage_names: List[str] = [w.name for w in self.lineage()]
        if lineage_names[-1] is None:  # root usually won't have a name
            lineage_names = lineage_names[:-1]

        r = list(reversed([self.name, *lineage_names]))
        return ".".join(r)

    def lineage(self) -> List["Wrapper"]:
        lineage: List[Wrapper] = []
        parent = self.parent
        while parent is not None:
            lineage.append(parent)
            parent = parent.parent
        return lineage

    @property
    def nesting_level(self) -> int:
        return len(self.lineage())
        level = 0
        parent = self.parent
        while parent is not None:
            parent = parent.parent
            level += 1
        return level

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @property
    @abstractmethod
    def parent(self) -> Optional["Wrapper"]:
        pass

    @property
    @abstractmethod
    def required(self) -> bool:
        pass

    @required.setter
    @abstractmethod
    def required(self, value: bool):
        pass

    @property
    @abstractmethod
    def field(self) -> Optional[Field]:
        pass

    @property
    @abstractmethod
    def type(self) -> Type:
        pass


class AggregateWrapper(Wrapper[T]):
    """Wrapper for types that have fields (i.e. Dataclasses and Choices)."""

    @abstractmethod
    def register_actions(self, parser: argparse.ArgumentParser) -> None:
        pass



"""
A Choice Type aka "Sum Type" is a type that can be one of several types. Typically this is through subtyping,
but could be a tagged union or something else.

Typically you'll write a config like this:

```yaml
model:
    type: gpt
    pdrop: 0.1
    layers: 3
```

with a class hierarchy that looks like this:

```python
class ModelConfig(ChoiceRegistry):
    pass

@ModelConfig.register_subclass("gpt")
class GPTConfig(ModelConfig):
    pdrop: float
    layers: int
```

For now, Draccus choice types are implemented as classes that conform to a protocol. The protocol defines
a "get_choice_class(name)" method that returns the class corresponding to the given name, along with a few
other methods.

We may add support for Unions.

All registered types must be *dataclasses*
"""

import functools
import importlib
import pkgutil
from typing import Any, Callable, ClassVar, Dict, Optional, Protocol, Type, TypeVar, overload, runtime_checkable

T = TypeVar("T")


CHOICE_TYPE_KEY = "type"
"""name of key to use in configuration to specify the type of a choice type"""


@runtime_checkable
class ChoiceType(Protocol):
    @classmethod
    def get_choice_class(cls, name: str) -> Any:
        ...

    @classmethod
    def get_known_choices(cls) -> Dict[str, Any]:
        ...

    @classmethod
    def get_choice_name(cls, subcls: Type) -> str:
        """
        Returns the name of the given subclass of this choice type.
        """
        ...

    @classmethod
    def default_choice_name(cls) -> Optional[str]:
        """
        Returns the name of the default subclass of this choice type.
        """
        ...


class ChoiceRegistryBase(ChoiceType):
    _choice_registry: ClassVar[Dict[str, Any]]

    @classmethod
    def get_choice_class(cls, name: str) -> Any:
        return cls._choice_registry[name]

    @classmethod
    def get_known_choices(cls) -> Dict[str, Any]:
        return cls._choice_registry

    @classmethod
    def get_choice_name(cls, subcls: Type) -> str:
        for name, choice_type in cls._choice_registry.items():
            if choice_type == subcls:
                return name
        raise ValueError(f"Cannot find choice name for {subcls}")

    @classmethod
    def default_choice_name(cls) -> Optional[str]:
        return None

    @overload
    @classmethod
    def register_subclass(cls, name: str, choice_type: Type) -> Type[T]:
        ...

    @overload
    @classmethod
    def register_subclass(cls, name: str) -> Callable[[Type[T]], Type[T]]:
        ...

    @classmethod
    def register_subclass(cls, name: str, choice_type: Optional[Type[T]] = None):
        if choice_type is None:
            return functools.partial(cls.register_subclass, name)

        if name in cls._choice_registry:
            other_choice_type = cls._choice_registry[name]
            if other_choice_type != choice_type:
                raise ValueError(
                    f"Cannot register {choice_type} as {name} because {other_choice_type} is already registered as"
                    f" {name}"
                )

        cls._choice_registry[name] = choice_type
        return choice_type


class ChoiceRegistry(ChoiceRegistryBase):
    _choice_registry: ClassVar[Dict[str, Any]]

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "_choice_registry"):
            cls._choice_registry: ClassVar[Dict[str, Any]] = {}


class PluginRegistry(ChoiceRegistryBase):
    """
    A ChoiceRegistry that allows for discovery of plugins.

    It follows the "namespace package plugin" pattern described here:
    https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/#using-namespace-packages

    To use this, you should put all your plugins in a single package, and then pass the path to that package
    to the constructor of this class. Then, when the class looks for plugins, it will look in that package and
    automatically import them.

    This is useful for things like models, where you want to be able to add new models without having to modify
    the code of the library itself.

    Usage:
    ```python
    @dataclasses.dataclass
    class ModelConfig(PluginRegistry):

        pass

    @dataclasses.dataclass
    class GPTConfig(ModelConfig):
        pass

    ModelConfig.register_subclass("gpt", GPTConfig)

    the syntax for a yaml file would be:
    ```yaml
    model:
      type: gpt
      <config for gpt>
    ```

    Unlike with ClassRegistry, import doesn't happen until you call get_choice_class or get_known_choices,
    and you can split your plugins across multiple files.
    """

    # TODO: is it better to lazily import plugins?
    _choice_registry: ClassVar[Dict[str, Any]]
    discover_packages_path: ClassVar[str]
    _did_discover_packages: ClassVar[bool]

    def __init_subclass__(cls, discover_packages_path: Optional[str] = None, **kwargs):
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "_choice_registry"):
            cls._choice_registry = {}
        if not hasattr(cls, "discover_packages_path"):
            if discover_packages_path is None:
                raise ValueError("discover_packages_path must be specified in the class or constructor")
            cls.discover_packages_path = discover_packages_path
        cls._did_discover_packages = False

    @classmethod
    def get_choice_class(cls, name: str) -> Any:
        cls._discover_packages()
        return cls._choice_registry[name]

    @classmethod
    def get_known_choices(cls) -> Dict[str, Any]:
        cls._discover_packages()
        return cls._choice_registry

    @classmethod
    def _discover_packages(cls):
        if cls._did_discover_packages:
            return

        # from https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/
        package_module = importlib.import_module(cls.discover_packages_path, __package__)

        def iter_namespace(ns_pkg):
            # Specifying the second argument (prefix) to iter_modules makes the
            # returned name an absolute name instead of a relative one. This allows
            # import_module to work without having to do additional modification to
            # the name.
            return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")

        for _finder, pkg_name, _ispkg in iter_namespace(package_module):
            importlib.import_module(pkg_name)
            # registration should happen in the initialization of the package, so importing is sufficient

        cls._did_discover_packages = True



import dataclasses

from tests.draccus_choice_plugins.model_config import ModelConfig


@ModelConfig.register_subclass("gpt")
@dataclasses.dataclass
class GptConfig(ModelConfig):
    attn_pdrop: float = 0.1



import sys
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Set, Tuple, Type

import pytest

from .testutils import ParsingError, TestSetup, format_list_using_brackets, parametrize, raises


def test_list_one_element(simple_attribute):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class Container(TestSetup):
        a: List[some_type] = field(default_factory=list)

    c = Container.setup("")
    assert c.a == []
    c = Container.setup(f"--a [{passed_value}]")
    assert c.a == [expected_value], Container.get_help_text()


@pytest.fixture
def ContainerClass():
    if sys.version_info < (3, 10):

        @dataclass
        class ContainerClass(TestSetup):
            a: Tuple[int, ...]
            b: List[int]
            c: Tuple[str, ...] = tuple()
            d: List[int] = field(default_factory=list)

    else:

        @dataclass
        class ContainerClass(TestSetup):
            a: tuple[int, ...]
            b: List[int]
            c: Tuple[str, ...] = tuple()
            d: list[int] = field(default_factory=list)

    return ContainerClass


def test_single_element_list(ContainerClass):
    container = ContainerClass.setup("--a [1] --b [4] --c [7] --d [10]")
    assert container.a == (1,)
    assert container.b == [4]
    assert container.c == ("7",)
    assert container.d == [10]


def test_required_attributes_works(ContainerClass):
    with raises(ParsingError):
        ContainerClass.setup("--b [4]")

    with raises(ParsingError):
        ContainerClass.setup("--a [4]")

    container = ContainerClass.setup("--a [4] --b [5]")
    assert container == ContainerClass(a=(4,), b=[5])


def test_default_value(ContainerClass):
    container = ContainerClass.setup("--a [1] --b '[4, 5, 6]'")
    assert container.a == (1,)
    assert container.b == [4, 5, 6]
    assert container.c == tuple()
    assert container.d == list()


@parametrize(
    "item_type, passed_values",
    [
        (int, [[1, 2], [4, 5], [7, 8]]),
        (float, [[1.1, 2.1], [4.2, 5.2], [7.2, 8.2]]),
        (str, [["a", "b"], ["c", "d"], ["e", "f"]]),
        (bool, [[True, True], [True, False], [False, True]]),
    ],
)
def test_parse_multiple_with_list_attributes(
    item_type: Type,
    passed_values: List[List[Any]],
):
    @dataclass
    class SomeClass(TestSetup):
        a: List[item_type] = field(default_factory=list)  # type: ignore
        """some docstring for attribute 'a'"""

    for value in passed_values:
        arguments = "--a " + format_list_using_brackets(value)
        result = SomeClass.setup(arguments)
        assert result == SomeClass(a=value)


@parametrize(
    "item_type, type_hint, value, arg",
    [
        (list, List, [1, 2, 3], "[1, 2, 3]"),
        (Optional[list], Optional[List], [1, 2, 3], "[1, 2, 3]"),
        (set, Set, {1, 2, 3}, "[1, 2, 3]"),
        (tuple, Tuple, (1, 2, 3), "[1, 2, 3]"),
        (dict, Dict, {1: 2}, "{1: 2}"),
    ],
)
def test_collection_no_type(item_type, type_hint, value, arg):
    @dataclass
    class ContainerHint(TestSetup):
        a: type_hint

    c = ContainerHint.setup(f"--a '{arg}'")
    assert c.a == value

    c = ContainerHint.setup(config=f"""a: {arg}""")
    assert c.a == value

    @dataclass
    class ContainerType(TestSetup):
        a: item_type

    c = ContainerType.setup(f"--a '{arg}'")
    assert c.a == value

    c = ContainerType.setup(config=f"""a: {arg}""")
    assert c.a == value


def test_list_of_dataclasses():
    @dataclass
    class Inner(TestSetup):
        a: float

    @dataclass
    class Outer(TestSetup):
        a: Optional[List[Inner]]

    c = Outer.setup("""--a '[{"a": 1}, {"a": 2}]'""")
    assert c.a == [Inner(a=1), Inner(a=2)]

    @dataclass
    class OptionalOuter(TestSetup):
        b: Optional[Outer] = None

    c = OptionalOuter.setup("""--b.a '[{"a": 1}, {"a": 2}]'""")
    assert c.b == Outer(a=[Inner(a=1), Inner(a=2)])

    yaml = """
b:
    a:
        - a: 1
        - a: 2
    """
    c = OptionalOuter.setup(config=yaml)

    assert c.b == Outer(a=[Inner(a=1), Inner(a=2)])



from abc import ABC, abstractmethod
from enum import Enum
from inspect import isclass


class Parser(ABC):
    @staticmethod
    @abstractmethod
    def parse_string(s):
        pass

    @staticmethod
    @abstractmethod
    def load_config(stream):
        pass

    @staticmethod
    @abstractmethod
    def save_config(d, stream=None, **kwargs):
        pass


class ParserEnum(Enum):
    def __init__(self, *args):
        if not isclass(self.value) or not issubclass(self.value, Parser):
            raise TypeError("All enum values must be parsers")


class YAMLParser(Parser):
    @staticmethod
    def parse_string(s):
        import yaml  # type: ignore

        from .yaml_loader import SafeLoaderWithInclusion

        return yaml.load(s, SafeLoaderWithInclusion)

    @staticmethod
    def load_config(stream):
        import yaml  # type: ignore

        from .yaml_loader import FullLoaderWithInclusion

        return yaml.load(stream, FullLoaderWithInclusion)

    @staticmethod
    def save_config(d, stream=None, **kwargs):
        import yaml

        return yaml.dump(d, stream, **kwargs)


class JSONParser(Parser):
    @staticmethod
    def parse_string(s):
        import json

        try:
            return json.loads(s)
        except json.decoder.JSONDecodeError:
            return s  # No parsing to be done, simply return value

    @staticmethod
    def load_config(stream):
        import json

        return json.load(stream)

    @staticmethod
    def save_config(d, stream=None, **kwargs):
        import json

        if stream is None:
            return json.dumps(d, **kwargs)
        else:
            return json.dump(d, stream, **kwargs)


class TOMLParser(Parser):
    @staticmethod
    def parse_string(s):
        import toml  # type: ignore

        try:
            return toml.loads(f"val = {s}")["val"]
        except toml.decoder.TomlDecodeError:
            return s

    @staticmethod
    def load_config(stream):
        import toml

        return toml.load(stream)

    @staticmethod
    def save_config(d, stream=None, **kwargs):
        import toml

        if stream is None:
            return toml.dumps(d, **kwargs)
        else:
            return toml.dump(d, stream, **kwargs)



import shlex
import sys
import tempfile
from contextlib import contextmanager, redirect_stderr
from io import StringIO
from typing import Any, Callable, List, Optional, Type, TypeVar

import pytest

import draccus
from draccus import ParsingError

xfail = pytest.mark.xfail
parametrize = pytest.mark.parametrize


def xfail_param(*args, reason: str):
    if len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]
    return pytest.param(*args, marks=pytest.mark.xfail(reason=reason))


Dataclass = TypeVar("Dataclass")


@contextmanager
def raises(exception=ParsingError, match=None, code: Optional[int] = None):
    with pytest.raises(exception, match=match):
        yield


@contextmanager
def exits_and_writes_to_stderr(match: str = ""):
    s = StringIO()
    with redirect_stderr(s), raises(SystemExit):
        yield
    s.seek(0)
    err_string = s.read()
    if match:
        assert match in err_string, err_string
    else:
        assert err_string, err_string


@contextmanager
def raises_missing_required_arg():
    with exits_and_writes_to_stderr("the following arguments are required"):
        yield


@contextmanager
def raises_expected_n_args(n: int):
    with exits_and_writes_to_stderr(f"expected {n} arguments"):
        yield


@contextmanager
def raises_unrecognized_args(*args: str):
    with exits_and_writes_to_stderr("unrecognized arguments: " + " ".join(args or [])):
        yield


def assert_help_output_equals(actual: str, expected: str):
    # Replace the start with `prog`, since the tests runner might not always be
    # `pytest`, could also be __main__ when debugging with VSCode
    prog = sys.argv[0].split("/")[-1]
    if prog != "pytest":
        expected = expected.replace("usage: pytest", f"usage: {prog}")
    # remove = string.punctuation + string.whitespace
    actual_str = "".join(actual.split())
    expected_str = "".join(expected.split())
    assert actual_str == expected_str, f"{actual_str} != {expected_str}"


T = TypeVar("T")


class TestSetup:
    @classmethod
    def setup(cls: Type[Dataclass], arguments: Optional[str] = "", config: Optional[str] = None) -> Dataclass:
        """Basic setup for a tests.

        Keyword Arguments:
            arguments {Optional[str]} -- The arguments to pass to the parser (default: {""})
            dest {Optional[str]} -- the attribute where the argument should be stored. (default: {None})

        Returns:
            {cls}} -- the class's type.
        """
        if arguments is not None:
            arguments = shlex.split(arguments)  # type: ignore
        if config is not None:
            f = tempfile.NamedTemporaryFile(suffix=".yaml")
            with open(f.name, "w") as fd:
                fd.write(config)
            cfg = draccus.parse(config_class=cls, args=arguments, prog="draccus", config_path=f.name)
        else:
            cfg = draccus.parse(config_class=cls, args=arguments, prog="draccus")
        return cfg

    @classmethod
    def get_help_text(
        cls,
    ) -> str:
        import contextlib
        from io import StringIO

        f = StringIO()
        with contextlib.suppress(SystemExit), contextlib.redirect_stdout(f):
            _ = cls.setup(
                "--help",
            )
        s = f.getvalue().strip()

        # python changed "optional arguments:" to "options" in 3.9
        import re

        replace_options = re.compile(r"(?m)^(optional arguments:|options:)$")
        s = replace_options.sub("options:", s)
        return s


ListFormattingFunction = Callable[[List[Any]], str]
ListOfListsFormattingFunction = Callable[[List[List[Any]]], str]


def format_list_using_spaces(value_list: List[Any]) -> str:
    return " ".join(str(p) for p in value_list)


def format_list_using_brackets(value_list: List[Any]) -> str:
    return f"[{','.join(str(p) for p in value_list)}]"


def format_list_using_single_quotes(value_list: List[Any]) -> str:
    return f"'{format_list_using_spaces(value_list)}'"


def format_list_using_double_quotes(value_list: List[Any]) -> str:
    return f'"{format_list_using_spaces(value_list)}"'


def format_lists_using_brackets(list_of_lists: List[List[Any]]) -> str:
    return " ".join(format_list_using_brackets(value_list) for value_list in list_of_lists)


def format_lists_using_double_quotes(list_of_lists: List[List[Any]]) -> str:
    return " ".join(format_list_using_double_quotes(value_list) for value_list in list_of_lists)


def format_lists_using_single_quotes(list_of_lists: List[List[Any]]) -> str:
    return " ".join(format_list_using_single_quotes(value_list) for value_list in list_of_lists)



from dataclasses import dataclass

from draccus.utils import DecodingError

from .testutils import *


def test_no_default_argument(simple_attribute):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type

    cfg = draccus.parse(config_class=SomeClass, args=shlex.split(f"--a {passed_value}"))
    assert cfg == SomeClass(a=expected_value)

    with raises(DecodingError):
        draccus.parse(config_class=SomeClass, args="")


def test_default_dataclass_argument(simple_attribute, silent):
    some_type, passed_value, expected_value = simple_attribute

    @dataclass
    class SomeClass:
        a: some_type = expected_value

    cfg = draccus.parse(config_class=SomeClass, args="")
    assert cfg == SomeClass(a=expected_value)



from dataclasses import dataclass, field
from pathlib import Path
from typing import Optional

import draccus

# logging.basicConfig(level=logging.DEBUG)


@dataclass
class ComputeConfig:
    """Config for training resources"""

    # The number of workers for training
    workers: int = field(default=8)
    # The number of workers for training
    eval_workers: Optional[int] = field(default=None)

    def __post_init__(self):
        # A builtin method of dataclasses, used for post-processing our configuration.
        self.eval_workers = self.eval_workers or self.workers


@dataclass
class OptimConfig(draccus.ChoiceRegistry):
    learning_rate: float = 0.001


@OptimConfig.register_subclass("adam")
class AdamConfig(OptimConfig):
    beta1: float = 0.9
    beta2: float = 0.999
    epsilon: float = 1e-8


@OptimConfig.register_subclass("sgd")
class SGDConfig(OptimConfig):
    momentum: float = 0.9
    nesterov: bool = False


@dataclass
class LogConfig:
    """Config for logging arguments"""

    # The experiment name
    exp_name: str = field(default="default_exp")
    # The experiment root folder path
    exp_root: Path = field(default=Path("/share/experiments"))

    @property
    def exp_dir(self) -> Path:
        # Properties are great for arguments that can be derived from existing ones
        return self.exp_root / self.exp_name


# TrainConfig will be our main configuration class.
# Notice that default_factory is the standard way to initialize a class argument in dataclasses


@dataclass
class TrainConfig:
    log: LogConfig = LogConfig()
    compute: ComputeConfig = ComputeConfig()
    optim: OptimConfig = AdamConfig()


@draccus.wrap()
def main(cfg: TrainConfig):
    print(f"Training {cfg.log.exp_name}...")
    print(f"\tUsing optimizer {cfg.optim}")
    print(f"\tUsing {cfg.compute.workers} workers and {cfg.compute.eval_workers} evaluation workers")
    print(f"\tSaving to {cfg.log.exp_dir}")
    print(draccus.dump(cfg))


if __name__ == "__main__":
    main()



from dataclasses import dataclass

import pytest

from draccus import ParsingError
from draccus.utils import DecodingError

from .testutils import TestSetup


@dataclass
class Base(TestSetup):
    """Some extension of base-class `Base`"""

    a: int = 5
    f: bool = False


def test_bool_attributes_work():
    true_strings = ["True", "true"]
    for s in true_strings:
        ext = Base.setup(f"--a 5 --f {s}")
        assert ext.f is True

    false_strings = ["False", "false"]
    for s in false_strings:
        ext = Base.setup(f"--a 5 --f {s}")
        assert ext.f is False


def test_bool_doesnt_parse_non_bools():
    with pytest.raises(DecodingError) as e:
        Base.setup("--a 5 --f 5")

    assert e.value.key_path == ("f",)

    with pytest.raises(DecodingError):
        Base.setup("--a 5 --f foo")



from dataclasses import dataclass

from draccus.wrappers.docstring import HelpOrder, get_attribute_docstring, get_preferred_help_text

from .testutils import TestSetup


@dataclass
class PreferredHelpConfig(TestSetup):
    """A simple base-class example"""

    # fmt: off
    a: int                                      # Parameter (int) controlling number of iterations

    # Cache Parameter
    cache: str                                  # Path to cache directory (as string)

    # Sets the mode of miscellaneous parameter handling
    misc: str = "standard"
    misc2: str = "left-first"                   # Controls the behavior of `misc2`
    # fmt: on


def test_get_preferred_help_text():
    a_docstring = get_attribute_docstring(PreferredHelpConfig, "a")
    a_help_inline = get_preferred_help_text(a_docstring, "inline")
    assert a_help_inline == "Parameter (int) controlling number of iterations"

    a_help_above = get_preferred_help_text(a_docstring, HelpOrder.above)
    assert a_help_above != "# fmt: off"

    # If "preferred" doesn't exist, goes through "default" order --> "inline" --> "above" --> "below"
    a_help_below = get_preferred_help_text(a_docstring, "below")
    assert a_help_below == a_help_inline

    misc2_docstring = get_attribute_docstring(PreferredHelpConfig, "misc2")
    misc2_help_inline = get_preferred_help_text(misc2_docstring, preferred_help="inline")
    assert misc2_help_inline == "Controls the behavior of `misc2`"
    assert misc2_docstring.comment_above == misc2_docstring.docstring_below == ""

    misc2_help_above = get_preferred_help_text(misc2_docstring, preferred_help="above")
    assert misc2_help_above == misc2_help_inline


def test_default_help_inline():
    target = """
usage: draccus [-h] [--config_path str] [--a int] [--cache str] [--misc str]
               [--misc2 str]

options:
  -h, --help         show this help message and exit
  --config_path str  Path for a config file to parse with draccus (default:
                     None)

PreferredHelpConfig:
  A simple base-class example

  --a int            Parameter (int) controlling number of iterations
                     (default: None)
  --cache str        Path to cache directory (as string) (default: None)
  --misc str         Sets the mode of miscellaneous parameter handling
                     (default: standard)
  --misc2 str        Controls the behavior of `misc2` (default: left-first)
    """
    assert PreferredHelpConfig.get_help_text().strip() == target.strip()



import argparse
import dataclasses
from logging import getLogger
from typing import Dict, List, Optional, Type, Union, cast

from draccus.utils import Dataclass, DataclassType

from .. import utils
from ..choice_types import ChoiceType
from ..parsers.decoding import has_custom_decoder
from . import docstring
from .field_wrapper import FieldWrapper
from .wrapper import AggregateWrapper, Wrapper

logger = getLogger(__name__)


class DataclassWrapper(AggregateWrapper[Type[Dataclass]]):
    def __init__(
        self,
        dataclass: Type[Dataclass],
        name: Optional[str] = None,
        # TODO(dlwh): they aren't using the defaults?
        default: Optional[Union[Dataclass, Dict]] = None,
        parent: Optional["Wrapper"] = None,
        _field: Optional[dataclasses.Field] = None,
        preferred_help: str = docstring.HelpOrder.inline,
    ):
        self.dataclass = dataclass
        self._name = name
        self.default = default

        self._required: bool = False
        self._explicit: bool = False
        self._children: List[Wrapper] = []
        self._parent = parent
        # the field of the parent, which contains this child dataclass.
        self._field = _field

        # preferred parse for docstring / help text in < inline | above | below >
        self.preferred_help = preferred_help

        # the default values
        self._defaults: List[Dataclass] = []

        if default:
            # TODO(dlwh): I don't like this
            self.defaults = [default]  # type: ignore

        self.optional: bool = False

        for field in dataclasses.fields(self.dataclass):  # type: ignore
            child = _wrap_field(self, field, preferred_help=self.preferred_help)
            if child is not None:
                self._children.append(child)

    def register_actions(self, parser: argparse.ArgumentParser) -> None:
        group = parser.add_argument_group(title=self.title, description=self.description)

        for child in self._children:
            if isinstance(child, AggregateWrapper):
                # Child name will always be populated as this is done via our code inside `_wrap_field`
                parser.add_argument("--" + child.name, type=str, required=False, help="Config file for " + child.name)
                child.register_actions(parser)
            elif isinstance(child, FieldWrapper):
                child.add_action(group)

    @property
    def name(self) -> str:
        # TODO(dlwh): I don't like this
        return self._name  # type: ignore

    @property
    def parent(self) -> Optional["Wrapper"]:
        return self._parent

    # @property
    # def defaults(self) -> List[Dataclass]:
    #     raise NotImplementedError("This should be overridden")
    #     if self._defaults:
    #         return self._defaults
    #     if self._field is None:
    #         return []
    #     assert self.parent is not None
    #     if self.parent.defaults:
    #         self._defaults = []
    #         for default in self.parent.defaults:
    #             if default is None:
    #                 default = None
    #             else:
    #                 default = getattr(default, self.name)
    #             self._defaults.append(default)
    #     else:
    #         try:
    #             default_field_value = utils.default_value(self._field)
    #         except TypeError as e:
    #             # utils.default_value tries to construct the field to get default value and might fail
    #             # if the field has some required arguments
    #             logger.debug(f"Could not get default value for field '{self._field.name}'\n\tUnderlying Error: {e}")
    #             default_field_value = dataclasses.MISSING
    #         if isinstance(default_field_value, _MISSING_TYPE):
    #             self._defaults = []
    #         else:
    #             self._defaults = [default_field_value]
    #     return self._defaults
    #
    # @defaults.setter
    # def defaults(self, value: List[Dataclass]):
    #     self._defaults = value

    @property
    def title(self) -> str:
        title = self.dataclass.__qualname__
        if self.dest is not None:  # Show name if exists
            title += f" ['{self.dest}']"
        return title

    @property
    def description(self) -> str:
        if self.parent and self._field:
            doc = docstring.get_attribute_docstring(self.parent.type, self._field.name)
            help_text = docstring.get_preferred_help_text(doc, preferred_help=self.preferred_help)
            if help_text is not None:
                return help_text
        class_doc = self.dataclass.__doc__ or ""
        if class_doc.startswith(f"{self.dataclass.__name__}("):
            return ""  # The base dataclass doc looks confusing, remove it
        return class_doc

    @property
    def required(self) -> bool:
        return self._required

    @required.setter
    def required(self, value: bool):
        self._required = value
        for child_wrapper in self._children:
            child_wrapper.required = value

    @property
    def field(self) -> Optional[dataclasses.Field]:
        return self._field

    @property
    def type(self) -> Type[Dataclass]:
        return self.dataclass


def _wrap_field(
    parent: Optional[Wrapper],
    field: dataclasses.Field,
    preferred_help: str = docstring.HelpOrder.inline,
) -> Optional[Wrapper]:
    if not field.init:
        return None

    elif has_custom_decoder(field.type):
        field_wrapper = FieldWrapper(field, parent=parent, preferred_help=preferred_help)
        logger.debug(f"wrapped field at {field_wrapper.dest} has a default value of {field_wrapper.default}")
        return field_wrapper
    elif utils.is_choice_type(field.type):
        from .choice_wrapper import ChoiceWrapper

        return ChoiceWrapper(
            cast(Type[ChoiceType], field.type), field.name, parent=parent, _field=field, preferred_help=preferred_help
        )

    elif utils.is_tuple_or_list_of_dataclasses(field.type):
        logger.debug(f"wrapped field at {field.name} is a list of dataclasses, treating a ordinary field for argparse")
        field_wrapper = FieldWrapper(field, parent=parent, preferred_help=preferred_help)
        return field_wrapper
        # raise NotImplementedError(
        #     f"Field {field.name} is of type {field.type}, which isn't supported yet. (container of a dataclass type)"
        # )

    elif dataclasses.is_dataclass(field.type):
        # handle a nested dataclass attribute
        dataclass, name = (cast(DataclassType, field.type)), field.name
        child_wrapper = DataclassWrapper(dataclass, name, parent=parent, _field=field, preferred_help=preferred_help)
        return child_wrapper

    elif utils.is_optional_or_union_with_dataclass_type_arg(field.type):
        # TODO(dlwh): I don't like this. Add UnionWrapper or something
        name = field.name
        from .choice_wrapper import UnionWrapper

        wrapper = UnionWrapper(field.type, name=name, parent=parent, _field=field, preferred_help=preferred_help)
        return wrapper

    else:
        # a normal attribute
        field_wrapper = FieldWrapper(field, parent=parent, preferred_help=preferred_help)
        logger.debug(f"wrapped field at {field_wrapper.dest} has a default value of {field_wrapper.default}")
        # self._children.append(field_wrapper)
        return field_wrapper



import yaml  # type: ignore
import yaml.representer  # type: ignore
import yamlinclude
from yaml import MappingNode
from yaml.constructor import ConstructorError  # type: ignore


class ConstructorWithGoodInclusion(yaml.constructor.SafeConstructor, yaml.representer.SafeRepresenter):
    def __init__(self):
        yaml.constructor.SafeConstructor.__init__(self)
        yaml.representer.SafeRepresenter.__init__(self)

    # this is a hack to get around the fact that yamlinclude doesn't work with the merge key <<
    def flatten_mapping(self, node):
        merge = []
        index = 0
        while index < len(node.value):
            key_node, value_node = node.value[index]
            if key_node.tag == "tag:yaml.org,2002:merge":
                del node.value[index]
                # this is the difference from the original method
                if value_node.tag == yamlinclude.YamlIncludeConstructor.DEFAULT_TAG_NAME:
                    value = self.construct_object(value_node)
                    value = self.represent_data(value)
                    if isinstance(value, MappingNode):
                        merge.extend(value.value)
                    else:
                        raise ConstructorError(
                            "while constructing a mapping",
                            node.start_mark,
                            "expected included node to be mapping for merging, but found %s" % value_node.id,
                            value_node.start_mark,
                        )
                elif isinstance(value_node, yaml.MappingNode):
                    self.flatten_mapping(value_node)
                    merge.extend(value_node.value)
                elif isinstance(value_node, yaml.SequenceNode):
                    submerge = []
                    for subnode in value_node.value:
                        if not isinstance(subnode, yaml.MappingNode):
                            raise ConstructorError(
                                "while constructing a mapping",
                                node.start_mark,
                                "expected a mapping for merging, but found %s" % subnode.id,
                                subnode.start_mark,
                            )
                        self.flatten_mapping(subnode)
                        submerge.append(subnode.value)
                    submerge.reverse()
                    for value in submerge:
                        merge.extend(value)
                else:
                    raise ConstructorError(
                        "while constructing a mapping",
                        node.start_mark,
                        "expected a mapping or list of mappings for merging, but found %s" % value_node.id,
                        value_node.start_mark,
                    )
            elif key_node.tag == "tag:yaml.org,2002:value":
                key_node.tag = "tag:yaml.org,2002:str"
                index += 1
            else:
                index += 1
        if merge:
            node.value = merge + node.value


class FullLoaderWithInclusion(yaml.FullLoader, ConstructorWithGoodInclusion):
    def __init__(self, *args, **kwargs):
        yaml.FullLoader.__init__(self, *args, **kwargs)
        ConstructorWithGoodInclusion.__init__(self)


class SafeLoaderWithInclusion(yaml.SafeLoader, ConstructorWithGoodInclusion):
    def __init__(self, *args, **kwargs):
        yaml.SafeLoader.__init__(self, *args, **kwargs)
        ConstructorWithGoodInclusion.__init__(self)


yamlinclude.YamlIncludeConstructor.add_to_loader_class(FullLoaderWithInclusion, relative=True)
yamlinclude.YamlIncludeConstructor.add_to_loader_class(SafeLoaderWithInclusion, relative=True)



import dataclasses


# ruff: noqa: A002
def field(
    *,
    default=dataclasses.MISSING,
    default_factory=dataclasses.MISSING,
    init=True,
    repr=True,
    hash=None,
    compare=True,
    metadata=None,
    is_mutable=False,
):
    if default is not dataclasses.MISSING and default_factory is not dataclasses.MISSING:
        raise ValueError("cannot specify both default and default_factory")
    if is_mutable and default is not dataclasses.MISSING:
        from copy import deepcopy

        copy_val = default

        def default_factory():
            return deepcopy(copy_val)

        default = dataclasses.MISSING

    return dataclasses.field(
        default=default,
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )
